# -----------------------------------------------------------------------------
#  macedon [CLI web service availability verifier]
#  (c) 2022-2023 A. Shavykin <0.delameter@gmail.com>
# -----------------------------------------------------------------------------
import re
import typing as t

import click
from requests.structures import CaseInsensitiveDict

from ._common import Task
from .logger import get_logger


class FileParser:
    def parse(self, file: click.File):
        try:
            yield from self._parse_jb_http_file(file)
        except Exception as e:
            get_logger().error(e)
        else:
            return
        try:
            yield from self._parse_plain(file)
        except Exception as e:
            get_logger().error(e)

    def _parse_plain(self, file: click.File) -> t.Iterable[Task]:
        while line := file.readline():
            try:
                yield Task(*self._extract_method_url(line))
            except ValueError as e:
                get_logger().error(e)
                continue

    def _parse_jb_http_file(self, file: click.File) -> t.Iterable[Task]:
        logger = get_logger()
        logger.debug("Attempting to parse jetbrains format")

        data = file.read()
        request_list = re.split(r"^###.*$\s*", data, flags=re.MULTILINE)
        request_filtered_list = [*filter(None, (r.strip() for r in request_list))]
        logger.debug(f"Detected {len(request_filtered_list)} requests")

        for idx, request in enumerate(request_filtered_list):
            logger.debug(f"Parsing request #{idx}")

            lines, last_empty_idx = self._filter_jb_http_file_lines(
                request.splitlines()
            )
            url, method = self._extract_method_url(lines[0])
            headers = CaseInsensitiveDict(
                {
                    k.strip(): v.strip()
                    for k, v in self._extract_headers(lines[1:last_empty_idx])
                }
            )
            body = None
            if last_empty_idx is not None:
                body_lines = lines[last_empty_idx:]
                body = "".join(body_lines)
                logger.debug(f"Found body ({len(body_lines)} lines, {len(body)} chars)")

            yield Task(url, method, headers, body)

    def _filter_jb_http_file_lines(
        self, lines: list[str]
    ) -> tuple[list[str], int | None]:
        last_empty_line_idx = None

        def _filter(line: str) -> bool:
            return not line.startswith("#")

        def _normalize(line: str, idx: int) -> str:
            nonlocal last_empty_line_idx
            if re.match("^\s*$", line):
                last_empty_line_idx = idx
                return ""
            return line

        filtered = [*filter(_filter, lines)]
        get_logger().debug(f"Filtered lines: {len(filtered)}")

        normalized = [_normalize(line, idx) for idx, line in enumerate(filtered)]
        get_logger().debug(f"Normalized lines: {len(normalized)}")
        get_logger().debug(f"Last empty line idx: {last_empty_line_idx}")

        return normalized, last_empty_line_idx

    def _extract_method_url(self, line: str) -> list[str, str]:
        if re.match("^([A-Z]+)\s+(https?://\S+$)", line):
            result = [*reversed(line.split(" ", 1))]
            get_logger().debug(f"Found method and url: {result}")
            return result
        raise ValueError(
            f"Invalid format, expected '{{method}} http(s)?://{{url}}', got: '{line}'"
        )

    def _extract_headers(self, lines: list[str]) -> t.Iterable[tuple[str, str]]:
        for line in lines:
            if not (m := re.match(r"^([a-zA-Z0-9_-]+):(.+)$", line)):
                continue
            header = m.group(1, 2)
            get_logger().debug(f"Found header: {header}")
            yield header


def get_parser():
    return _parser


def init_parser():
    global _parser
    _parser = FileParser()
    return _parser


_parser: FileParser | None = None



# -----------------------------------------------------------------------------
#  macedon [CLI web service availability verifier]
#  (c) 2022-2023 A. Shavykin <0.delameter@gmail.com>
# -----------------------------------------------------------------------------
import logging
import os
import sys
from logging import LogRecord
from os import P_PID

import pytermor as pt

from ._common import Options
from .io import get_stderr

VERBOSITY_LOG_LEVELS = {
    0: logging.CRITICAL,
    1: logging.INFO,
    2: logging.DEBUG,
}


class SgrFormatter(logging.Formatter):
    def format(self, record: LogRecord) -> str:
        result = super().format(record)
        fmt = pt.NOOP_STYLE
        match record.levelno:
            case logging.DEBUG:
                fmt = pt.cv.CYAN
            case logging.INFO:
                fmt = pt.cv.WHITE
            case logging.WARNING:
                fmt = pt.cv.YELLOW
            case logging.ERROR:
                fmt = pt.cv.RED
        return get_stderr().render(result, fmt)


def get_logger():
    return _logger


def init_loggers(options: Options):
    log_level = VERBOSITY_LOG_LEVELS.get(options.verbose, logging.WARNING)

    if options.verbose >= 2:
        logging.getLogger().setLevel(log_level)  # root logger

    from . import APP_NAME

    global _logger
    _logger = logging.getLogger(APP_NAME)
    _handler = logging.StreamHandler(sys.stderr)
    _handler.setFormatter(
        SgrFormatter("[%(levelname)-5.5s][%(name)s:%(threadName)s] %(message)s")
    )
    _handler.setLevel(log_level)
    _logger.addHandler(_handler)
    _logger.setLevel(log_level)

    pt_logger = logging.getLogger("pytermor")
    handler = logging.StreamHandler(sys.stderr)
    formatter = logging.Formatter(
        "[%(levelname)-5.5s][%(name)s|%(module)s] %(message)s"
    )
    handler.setFormatter(formatter)
    handler.setLevel(log_level)
    pt_logger.addHandler(handler)
    pt_logger.setLevel(log_level)

    return _logger


_logger: logging.Logger | None = None



# -----------------------------------------------------------------------------
#  macedon [CLI web service availability verifier]
#  (c) 2022-2023 A. Shavykin <0.delameter@gmail.com>
# -----------------------------------------------------------------------------
import threading as t
import time
from queue import Queue, Empty

import pytermor as pt
import requests
import urllib3.exceptions
from requests import Response

from ._common import Options, Task, SharedState
from .logger import get_logger
from .printer import get_printer


class Worker(t.Thread):
    def __init__(
        self,
        options: Options,
        task_pool: Queue[Task],
        idx: int,
        shared_state: SharedState,
    ):
        self._options = options
        self._task_pool = task_pool
        self._idx = idx
        self._shared_state = shared_state
        super().__init__(target=self.run, name=f"#{idx}")

    def run(self):
        logger = get_logger()
        printer = get_printer()

        while True:
            try:
                task = self._task_pool.get_nowait()
            except Empty:
                logger.debug(f"Empty queue, terminating")
                self._update_state("dead")
                return
            if not task:
                continue
            if (delay := self._options.delay) > 0:
                self._update_state("sleep")
                time.sleep(delay)

            response = None
            request_id = self._shared_state.last_request_id.next()
            request_params = dict(
                headers=task.headers,
                data=task.body,
                allow_redirects=True,
                timeout=self._options.timeout,
                verify=task.url.startswith("https"),
            )
            logger.info(
                f"Performing request #{request_id}: {task.method} {task.url} {request_params}"
            )
            exception = None
            self._update_state("request")
            time_before = time_after = time.time_ns()
            try:
                response = requests.request(task.method, task.url, **request_params)
            except urllib3.exceptions.InsecureRequestWarning as e:
                logger.warning(e)
            except requests.exceptions.RequestException as e:
                time_after = time.time_ns()
                logger.exception(e, exc_info=False)
                exception = e

            if response is not None:
                if response.ok:
                    self._shared_state.requests_success.next()
                else:
                    self._shared_state.requests_failed.next()

                self._shared_state.requests_latency.append(response.elapsed.total_seconds())
                printer.print_response(task, response, request_id)
                self._dump_response(response, request_id)
            else:
                self._shared_state.requests_failed.next()
                printer.print_failed_request(task, time_after - time_before, request_id, exception)
                logger.info(f"No response for #{request_id}")

    def _update_state(self, state: str):
        get_logger().debug(f"State -> {state}")
        self._shared_state.worker_states[self._idx] = state

    def _dump_response(self, response: Response, request_id: int):
        logger = get_logger()
        logger.info(
            f"Response #{request_id} metadata: {response.status_code} {response.headers}"
        )
        try:
            logger.debug(
                pt.dump(response.content.decode(), f"Response #{request_id} content")
            )
        except UnicodeError:
            logger.debug(
                pt.BytesTracer().apply(
                    response.content,
                    pt.TracerExtra(f"Response #{request_id} raw content"),
                )
            )



# -----------------------------------------------------------------------------
#  macedon [CLI web service availability verifier]
#  (c) 2023 A. Shavykin <0.delameter@gmail.com>
# -----------------------------------------------------------------------------
import re
import sys
import threading
from datetime import timedelta

import pytermor as pt
import requests
from pytermor import RT, Fragment

from ._common import Task, SharedState, Options
from .io import get_stdout
from .logger import get_logger


class Printer:
    COLUMN_PAD = 2
    CW_STATUS = 4
    CW_SIZE = 7
    CW_ELAPSED = 7

    SUCCESS_ST = pt.Style(fg=pt.cv.GREEN, bold=True)
    FAILURE_ST = pt.Style(fg=pt.cv.RED, bold=True)
    ERROR_ST = pt.Style(fg=pt.cv.RED, dim=True)
    REQUEST_ID_ST = pt.Style(fg=pt.cv.YELLOW)
    NO_VAL_ST = pt.Style(fg=pt.cv.GRAY_23)
    METHOD_ST = pt.Style(bold=True)

    def __init__(self, options: Options, shared_state: SharedState):
        self._options = options
        self._shared_state = shared_state
        self._lock = threading.Lock()
        self._request_table = pt.SimpleTable(sep=pt.pad(self.COLUMN_PAD))
        self._progress_table = pt.SimpleTable(sep="")

        self._size_formatter = None
        self._elapsed_formatter = None
        self._request_id_formatter = None
        self._progress_formatter = None

    def print_prolog(self):
        req_total = self._shared_state.requests_total.value
        threads = self._options.threads
        self._print_row(
            pt.Text(width=2),
            pt.Text(f"Threads:", width=12),
            pt.Text(str(threads), pt.Style(bold=True), width=6, align="right"),
        )
        self._print_row(
            pt.Text(width=2),
            pt.Text(f"Requests:", width=12),
            pt.Text(str(req_total), pt.Style(bold=True), width=6, align="right"),
        )
        self._print_separator()

    def print_response(self, task: Task, response: requests.Response, request_id: int):
        size = 0
        try:
            size = len(response.content)
        except Exception:
            pass

        self._lock.acquire()
        self._print_request_result(
            self._format_status_code(response),
            self._format_size(size),
            self._format_elapsed(response.elapsed),
            self._format_request_id(request_id),
            self._format_url(task.url, task.method, None),
        )
        self._print_progress()
        self._shared_state.requests_printed.next()
        self._lock.release()

    def print_failed_request(
        self, task: Task, time_ns: int | float, request_id: int, exception: Exception
    ):
        self._lock.acquire()
        self._print_request_result(
            self._format_error(exception),
            self._format_elapsed(time_ns),
            self._format_request_id(request_id),
            self._format_url(task.url, task.method, exception),
        )
        self._print_progress()
        self._shared_state.requests_printed.next()
        self._lock.release()

    def print_epilog(self, time_delta_ns: int):
        req_total = self._shared_state.requests_total.value
        req_success = self._shared_state.requests_success.value
        req_failed = self._shared_state.requests_failed.value
        success_st = self.SUCCESS_ST if req_success == req_total else None
        failed_st = self.FAILURE_ST if req_failed > 0 else None
        avg_latency = pt.utilmisc.median(self._shared_state.requests_latency)

        self._reset_cursor_x()
        self._print_separator()
        self._print_row(
            pt.Text(width=2),
            pt.Text(f"Successful:", width=12),
            pt.Text(str(req_success), success_st, width=6, align="right"),
        )
        self._print_row(
            pt.Text(width=2),
            pt.Text(f"Failed:", width=12),
            pt.Text(str(req_failed), failed_st, width=6, align="right"),
            pt.Fragment(f"  ({100*req_failed/req_total:.1f}%)"),
        )
        self._print_row(
            pt.Text(width=2),
            pt.Text(f"Avg time:", width=12),
            pt.Text(width=1),
            self._format_elapsed(timedelta(seconds=avg_latency)),
        )
        self._print_row(
            pt.Text(width=2),
            pt.Text(f"Total time:", width=12),
            pt.Text(width=1),
            self._format_elapsed(time_delta_ns),
        )

    def _print_request_result(self, *vals: pt.IRenderable | None):
        self._reset_cursor_x()

        result = self._request_table.pass_row(*filter(None, vals))
        get_stdout().echo_rendered(result)

    def _print_progress(self):
        self._print_row(
            pt.Text("[", width=3, align="center"),
            self._format_progress(),
            self._format_request_count(),
            pt.Text("]", width=3, align="center"),
            newline=False,
        )

    def _print_row(self, *vals: pt.IRenderable, newline: bool = True):
        stdout = get_stdout()
        result = self._progress_table.pass_row(*vals)
        stdout.echo(stdout.render(result), newline=newline)

    def _print_separator(self):
        self._print_row(pt.Text(width=25, fill="-"))

    def _reset_cursor_x(self):
        get_stdout().echo(pt.ansi.make_set_cursor_x_abs(1).assemble(), newline=False)

    def _get_output_mode(self, opt_color: bool | None) -> pt.OutputMode:
        if opt_color is None:
            return pt.OutputMode.AUTO
        if opt_color:
            return pt.OutputMode.TRUE_COLOR
        return pt.OutputMode.NO_ANSI

    @property
    def _is_color_allowed(self) -> bool:
        return get_stdout().renderer.is_format_allowed

    def _get_max_req_id_length(self) -> int:
        return len(str(self._shared_state.requests_total.value))

    def _format_no_val(self, width: int) -> pt.Text:
        return pt.Text("---", self.NO_VAL_ST, width=width, align="center")

    def _format_status_code(self, response: requests.Response) -> pt.Text:
        string = str(response.status_code)
        fmt = self.SUCCESS_ST if response.ok else self.FAILURE_ST
        return pt.Text(string, fmt, width=self.CW_STATUS, align="right")

    def _format_error(self, exception: Exception) -> pt.Text:
        return pt.Text(
            self._get_error_type(exception),
            self.FAILURE_ST,
            width=self.CW_STATUS + self.CW_SIZE,
            align="right",
        )

    def _format_size(self, size: int) -> pt.Text:
        if not self._size_formatter:
            self._size_formatter = pt.StaticBaseFormatter(
                max_value_len=3,
                unit="b",
                unit_separator="",
                allow_negative=False,
                discrete_input=True,
                pad=True,
                color=self._is_color_allowed,
            )
        return self._size_formatter.format(size)

    def _format_elapsed(self, elapsed: timedelta | int | float) -> RT:
        if not self._elapsed_formatter:
            self._elapsed_formatter = pt.StaticBaseFormatter(
                max_value_len=3,
                allow_negative=False,
                unit="s",
                unit_separator="",
                pad=True,
                color=self._is_color_allowed,
            )
        if isinstance(elapsed, (int, float)):
            seconds = elapsed / 1e9
        elif isinstance(elapsed, timedelta):
            seconds = elapsed.total_seconds()
        else:
            get_logger().error(f"Invalid type of 'elazpsed' metric: {elapsed!r}")
            return self._format_no_val(width=self.CW_ELAPSED)
        return self._elapsed_formatter.format(seconds)

    def _format_request_id(self, request_id: int) -> pt.Text:
        if not self._options.show_id:
            return pt.Text(width=0)

        label = Fragment("#")
        result = Fragment(f"{request_id:>d}", self.REQUEST_ID_ST)
        max_width = self._get_max_req_id_length() + len(str(label.string))
        return pt.Text(label, result, width=max_width, align="right")

    def _format_request_count(self) -> pt.Text:
        current = self._shared_state.requests_printed.value + 1
        total = self._shared_state.requests_total.value
        max_id_width = self._get_max_req_id_length()
        label = " "
        result = f"{label}{current:>{max_id_width}d}/{total:<{max_id_width}d}"
        return pt.Text(result, width=max_id_width * 2 + len(str(label)) + 1)

    def _format_progress(self) -> pt.Text:
        if not self._progress_formatter:
            self._progress_formatter = pt.StaticBaseFormatter(
                max_value_len=3,
                allow_negative=False,
                allow_fractional=False,
                unit_separator="",
                unit="%",
                pad=True,
                prefixes=[None, ""],
                color=self._is_color_allowed,
            )
        original_val = (
            100
            * (self._shared_state.requests_printed.value + 1)
            / self._shared_state.requests_total.value
        )
        if original_val <= 1:
            original_val = 0.00
        result = self._progress_formatter.format(original_val)
        return result

    def _format_url(
        self, url: str, method: str, exception: Exception | None
    ) -> pt.Text:
        method_len = max(len(m) for m in self._shared_state.methods)
        error = pt.Fragment(self._get_error_msg(exception), self.ERROR_ST)
        result = [
            pt.Fragment(f"{method:>{method_len}.{method_len}s} ", self.METHOD_ST),
            pt.Fragment(url + pt.pad(2)),
            error,
        ]
        return pt.Text(*result)

    def _get_error_type(self, exception: Exception | None) -> str:
        if not exception:
            return ""
        return str(exception.__class__.__qualname__)

    def _get_error_msg(self, exception: Exception | None) -> str:
        if not self._options.show_error:
            return ""
        if not exception:
            return ""
        if "Errno" in (exception_str := str(exception)):
            return re.search(r"\[Errno -?\d+][^)\']+", exception_str).group() or ""
        return ""


def get_printer() -> Printer:
    return _printer


def init_printer(options: Options, shared_state: SharedState) -> Printer:
    global _printer
    _printer = Printer(options, shared_state)

    return _printer


_printer: Printer | None = None



# -----------------------------------------------------------------------------
#  macedon [CLI web service availability verifier]
#  (c) 2022-2023 A. Shavykin <0.delameter@gmail.com>
# -----------------------------------------------------------------------------
import threading
import time
from queue import Queue

import click
import pytermor as pt

from ._common import Options, Task, SharedState
from .fileparser import get_parser
from .logger import get_logger
from .printer import get_printer
from .worker import Worker


class Synchronizer:
    def __init__(self, url_args: tuple[str], file: tuple[click.File], options: Options, shared_state: SharedState):
        self._options = options
        self._shared_state = shared_state
        self._task_pool = Queue[Task]()
        self._workers = []

        self._init_task_queue(url_args, file)
        self._init_workers()

    def run(self):
        printer = get_printer()
        printer.print_prolog()
        time_before = time_after = time.time_ns()

        for worker in self._workers:
            get_logger().debug(f"Starting worker {worker}")
            worker.start()
        for worker in self._workers:
            worker.join()

        time_after = time.time_ns()
        self._shared_state.requests_latency.sort()
        printer.print_epilog(time_after - time_before)


    def _init_task_queue(self, url_args: tuple[str], file: tuple[click.File]):
        for file_inst in file:
            try:
                for task in get_parser().parse(file_inst):
                    self._append_task(task)
            except Exception as e:
                get_logger().error(f"Failed to parse the file '{file}': {e}")
        if self._shared_state.requests_total.value > 0:
            return

        for url in url_args:
            if not url.startswith("http"):
                url = f"http://{url}"
            self._append_task(Task(url))

    def _append_task(self, task: Task):
        self._shared_state.methods.add(task.method)
        for _ in range(self._options.amount):
            self._task_pool.put_nowait(task)
            self._shared_state.requests_total.next()

    def _init_workers(self):
        self._shared_state.worker_states = ["init"]*self._options.threads
        for idx in range(self._options.threads):
            self._workers.append(
                Worker(self._options, self._task_pool, idx, self._shared_state)
            )



# -----------------------------------------------------------------------------
#  macedon [CLI web service availability verifier]
#  (c) 2023 A. Shavykin <0.delameter@gmail.com>
# -----------------------------------------------------------------------------
from __future__ import annotations
import sys

import click
import pytermor as pt
import typing as t

from ._common import Options

_stdout: IoProxy | None = None
_stderr: IoProxy | None = None


def get_stdout(require=True) -> IoProxy | None:
    global _stdout
    if not _stdout:
        if require:
            raise RuntimeError("Stdout proxy is uninitialized")
        return None
    return _stdout


def get_stderr(require=True) -> IoProxy | None:
    global _stderr
    if not _stderr:
        if require:
            raise RuntimeError("Stderr proxy is uninitialized")
        return None
    return _stderr


def init_io(options: Options) -> tuple[IoProxy, IoProxy]:
    global _stdout, _stderr
    if _stdout:
        raise RuntimeError("Stdout proxy is already initialized")
    if _stderr:
        raise RuntimeError("Stderr proxy is already initialized")

    _stdout = IoProxy(options.color, sys.stdout)
    _stderr = IoProxy(options.color, sys.stderr)
    pt.RendererManager.set_default(_stdout.renderer)
    return _stdout, _stderr


class IoProxy:
    def __init__(self, color_option: bool, io: t.IO):
        self._color = color_option
        self._renderer = self._make_renderer(io)
        self._io = io
        self._is_stderr = io == sys.stderr
        self._broken = False
        self._click_available = False

        try:
            import click

            self._click_available = isinstance(click.echo, t.Callable)
        except ImportError:
            pass

    @property
    def renderer(self) -> pt.IRenderer:
        return self._renderer

    def render(self, string: pt.RT = "", fmt: pt.FT = None) -> str:
        return pt.render(string, fmt, self._renderer, no_log=self._is_stderr)

    def echo_rendered(self, string: pt.RT = "", fmt: pt.FT = None):
        self.echo(self.render(string, fmt))

    def echo(self, string: str = "", newline: bool = True):
        try:
            if self._click_available:
                click.echo(string, file=self._io, color=self._color, nl=newline)
            else:
                print(
                    string,
                    file=self._io,
                    end=("\n" if newline else ""),
                    flush=not bool(string),
                )
        except BrokenPipeError:
            self._broken = True
            self._pacify_flush_wrapper()

    def _make_renderer(self, io: t.IO) -> pt.IRenderer:
        if self._color is False:
            return pt.renderer.NoOpRenderer()
        return pt.SgrRenderer(self._output_mode, io)

    def _pacify_flush_wrapper(self):
        sys.stdout = t.cast(t.TextIO, click.utils.PacifyFlushWrapper(sys.stdout))
        sys.stderr = t.cast(t.TextIO, click.utils.PacifyFlushWrapper(sys.stderr))

    @property
    def _output_mode(self) -> pt.OutputMode:
        if self._color is None:
            return pt.OutputMode.AUTO
        if self._color:
            return pt.OutputMode.TRUE_COLOR
        return pt.OutputMode.NO_ANSI



# -----------------------------------------------------------------------------
#  macedon [CLI web service availability verifier]
#  (c) 2022-2023 A. Shavykin <0.delameter@gmail.com>
# -----------------------------------------------------------------------------
import os
import sys

import click
import pytermor as pt
import urllib3
from click import Context, HelpFormatter
from urllib3.exceptions import InsecureRequestWarning

from . import APP_NAME, APP_VERSION
from ._common import Options, SharedState
from .fileparser import init_parser
from .io import init_io
from .logger import init_loggers, get_logger
from .printer import init_printer
from .synchronizer import Synchronizer


def invoke():
    callback()


class ClickCommand(click.Command):
    def format_usage(self, ctx: Context, formatter: HelpFormatter) -> None:
        pieces = self.collect_usage_pieces(ctx)
        formatter.write_usage(ctx.command_path, "|".join(pieces))


@click.command(
    cls=ClickCommand,
    no_args_is_help=True,
    context_settings=dict(max_content_width=pt.get_preferable_wrap_width()),
)
@click.argument("ENDPOINT_URL", type=str, nargs=-1)
@click.option(
    "-T",
    "--threads",
    type=int,
    default=1,
    show_default=True,
    help="Number of threads for concurrent request making.",
)
@click.option(
    "-n",
    "--amount",
    type=int,
    default=1,
    show_default=True,
    help="How many times each request will be performed.",
)
@click.option(
    "-d",
    "--delay",
    type=float,
    default=0,
    show_default=True,
    help="Seconds to wait between each request.",
)
@click.option(
    "-t",
    "--timeout",
    type=float,
    default=10,
    show_default=True,
    help="Seconds to wait for the response.",
)
@click.option(
    "-f",
    "--file",
    multiple=True,
    type=click.types.File(),
    help="Execute request(s) from a specified file. The file should contain a list of "
    "endpoints in the format '{method} {url}', one per line. Another supported "
    "(partially) format is JetBrains HTTP Client format, which additionally allows to "
    "specify request headers and body. The option can be specified multiple times. "
    "The ENDPOINT_URL argument(s) are ignored if this option is present.",
)
@click.option(
    "-c/-C",
    "--color/--no-color",
    is_flag=True,
    default=None,
    help="",
)
@click.option(
    "--show-id",
    is_flag=True,
    help="",
)
@click.option(
    "--show-error",
    is_flag=True,
    help="",
)
@click.option(
    "-v",
    "--verbose",
    count=True,
    type=click.types.IntRange(min=0, max=2, clamp=True),
    default=0,
    help="Display detailed info on every request.",
)
@click.pass_context
def callback(
    ctx: click.Context, endpoint_url: tuple[str], file: tuple[click.File], **kwargs
):
    options = Options(**kwargs)
    shared_state = SharedState()
    urllib3.disable_warnings(InsecureRequestWarning)

    init_io(options)
    init_loggers(options)
    _log_init_info(options)

    init_parser()
    init_printer(options, shared_state)
    sync = Synchronizer(endpoint_url, file, options, shared_state)
    sync.run()


def _log_init_info(options: Options):
    logger = get_logger()
    logger.debug(
        f"{APP_NAME} {APP_VERSION} "
        f"PID={os.getpid()} PPID={os.getppid()} "
        f"UID={os.getuid()} CWD={os.getcwd()}"
    )
    logger.debug(f"Args: {sys.argv!r}")
    logger.debug(f"Options: {options!r}")



# -----------------------------------------------------------------------------
#  macedon [CLI web service availability verifier]
#  (c) 2023 A. Shavykin <0.delameter@gmail.com>
# -----------------------------------------------------------------------------
import queue
from dataclasses import dataclass, field
from threading import Lock

from requests.structures import CaseInsensitiveDict


class ThreadSafeCounter:
    def __init__(self):
        self._value = 0
        self._lock = Lock()

    def next(self) -> int:
        with self._lock:
            self._value += 1
            return self._value

    @property
    def value(self) -> int:
        return self._value


@dataclass(frozen=True)
class Options:
    color: bool
    threads: int
    amount: int
    delay: float
    timeout: float
    verbose: int
    show_id: bool
    show_error: bool


@dataclass(frozen=True)
class Task:
    url: str
    method: str = 'GET'
    headers: CaseInsensitiveDict = None
    body: str = None


@dataclass
class SharedState:
    last_request_id = ThreadSafeCounter()
    requests_total = ThreadSafeCounter()
    requests_printed = ThreadSafeCounter()
    requests_success = ThreadSafeCounter()
    requests_failed = ThreadSafeCounter()
    requests_latency: list[float] = field(default_factory=list)
    methods = set()
    worker_states: list[str] = None
